program hfb_f90
!  Hartree-Fock-Bogoliubov in an orbital basis  
!  Program written by G.F. Bertsch and A. Luo
!  newton algorithm for chemical potentials by Y. Alhassid 4/2008
!  To be distributed under the GNU license 
!  code description is at the top of the subroutines
!  simplified HFB block structure
        include 'variables.dat'
	integer :: i,j,k,l,stabilze_it
        real ::  scale,testp2
	real ::  Ncal, Zcal, lamNub, lamNlb, lamPub, lamPlb
	integer :: testp1, testp3
        real    :: x(2)
        logical :: check

	call input_data
!  wildenthal scaling for sd shell
        scale= (18./(16.+npart))**0.3
!!!        vw=vw*scale
  	call block_structure
	call form_q0  !  quadruple matrix q0 

! construct one-body part of the Fock matrix h1
        call form_h1
! construct the 2-body interaction in the orbital basis
	call vijkl

! construct the hfb block structure 
	call hfb_block_structure 
	if(initial.eq.0) then
	  i=0
	  ! initial Fock matrix is random symmetric matrix
	  call hfb_init_fock_deltam ! tested
	  ! construct the hfb matrix
	  call hfb_construct ! tested
          ! perform diagonalization
	  call hfb_diag1 ! tested
	  ! form rho and kappa
	  call new_rho_kappa_e
	  call hfb_get_energy	  
        else
          call read_rho
	  call read_kappa ! from fort.14
        endif
	
	
	! construct fock and deltam
	call fock_general  ! h without the chemical potential
	call new_delta
        call hfb_get_energy	
        call find_numbers(Ncal,Zcal)
        print*, 'numbers', Ncal,Zcal

	do l=1,iterN
	 oldrho = rho
	 oldkappa = kappa
	 if(flag_lambda .eq. 0)then
!	 call find_mu_bisection
         call newt(x,2,check)
           if(check) then
            print *,'Local minimum. Try another initial guess for lambda'
           end if
         lambdaN=x(1)
         lambdaP=x(2)
         else
           call hfb_construct
           call hfb_diag1
           call new_rho_kappa_e
           if(mod(npart,2) .ne. 0 )then
             call new_rho_kappa_o(block_nodd)
           endif
           if(mod(ppart,2) .ne. 0 )then
             call new_rho_kappa_o(block_podd)
           endif 
	 endif
!	 rho = rho * 0.5 + newrho * 0.5
 	 call test_convergence
         call Fock_general
	 call new_delta 
	 call hfb_get_energy
         call find_numbers(Ncal,Zcal)
         print*, 'it, Ncal,Zcal',l,Ncal,Zcal 
         print*, 'lambdaN,lambdaP',lambdaN,lambdaP
         print*, 'convergerho,kappa',converge_rho,converge_kappa
        enddo

!	call hfb_get_energy
!	print*, 'it, Ncal,Zcal',l,Ncal,Zcal
!       print*, 'lambdaN,lambdaP',lambdaN,lambdaP
!        print*, 'convergerho,kappa',converge_rho,converge_kappa
	
	
	
!        call find_exp_q0
        
        call write_rho
	call write_kappa


end program hfb_f90

subroutine test_convergence
include 'variables.dat'
integer :: i,j,k
converge_rho = 0.0
converge_kappa = 0.0
do i=1,nblock
 do j=1,ndb(i)
  do k=1,ndb(i)
   converge_rho = converge_rho + abs(oldrho(j,k,i) - rho(j,k,i))
 converge_kappa = converge_kappa + abs(oldkappa(j,k,i) - kappa(j,k,i))
  enddo
 enddo
enddo
end subroutine test_convergence
 
subroutine funcv(n,x,fvec)
 include 'variables.dat'
 integer   n
 real  ::  x(n), fvec(n)
 real  ::  Ncal, Zcal

  lambdaN=x(1)
  lambdaP=x(2)
  call hfb_construct ! here we add the chemical potential
  call hfb_diag1
  call new_rho_kappa_e
  if(mod(npart,2) .ne. 0 )then
   call new_rho_kappa_o(block_nodd)
  endif
  if(mod(ppart,2) .ne. 0 )then
   call new_rho_kappa_o(block_podd)
  endif

  call find_numbers(Ncal,Zcal)
  fvec(1)=Ncal-npart
  fvec(2)=Zcal-ppart

  return
end subroutine funcv

subroutine find_mu_bisection
include 'variables.dat'
integer :: i,j,k,stabilize_it
real ::  Ncal, Zcal, lamNub, lamNlb, lamPub, lamPlb 
 stabilize_it=30
 lamNlb=lamN_min
 lamNub=lamN_max
 lamPlb=lamP_min
 lamPub=lamP_max

 do j=1,stabilize_it
  lambdaN=(lamNub + lamNlb)/2.0
  lambdaP=(lamPub + lamPlb)/2.0
  call hfb_construct ! here we add the chemical potential
  call hfb_diag1
  call new_rho_kappa_e
  if(mod(npart,2) .ne. 0 )then
   call new_rho_kappa_o(block_nodd)
  endif
  if(mod(ppart,2) .ne. 0 )then
   call new_rho_kappa_o(block_podd)
  endif

  call find_numbers(Ncal,Zcal)
   if(Ncal.gt.(npart - 1.0e-7)) then
    lamNub=lambdaN
   else
    lamNlb=lambdaN
   endif
   if(Zcal.gt.(ppart - 1.0e-7)) then
    lamPub=lambdaP
   else
    lamPlb=lambdaP
   endif
  enddo
end subroutine find_mu_bisection
  
!subroutine find_mu_bisection_2
!include 'variables.dat'
!integer :: i,j,k,stabilize_it
!real ::  Ncal, Zcal
! stabilize_it=30

 

subroutine write_rho
include 'variables.dat'
integer :: i,j,k
  open(13)
  do i=1,nblock
    do j=1,ndb(i)
      write(13,12) (rho(k,j,i),k=1,ndb(i))
12 format(8f10.5)
    enddo
  enddo
  close(13)
end subroutine write_rho

subroutine read_rho
include 'variables.dat'
integer :: i,j,k
open(12)
  do i=1,nblock
    do j=1,ndb(i)
      read(12,*) (rho(k,j,i),k=1,ndb(i))
    enddo
  enddo
close(12)
end subroutine read_rho

subroutine find_exp_q0
 include 'variables.dat'
 ! using tr(Q rho) to get <Q>
 integer:: i,j,k,l
 
 expectQ0 = 0.0
 do i=1,nblock
  do j=1,ndb(i)
   do k=1,ndb(i)
     expectQ0 = expectQ0 + q0(j,k,i)* rho(k,j,i)
   enddo
  enddo
 enddo
end subroutine 



subroutine input_data
 include 'variables.dat'
        open(22,file='input.dat')
        read(22,*) npart, ppart, iterN, seed, lambda, alpha,initial, flag_lambda
        write(6,19) ppart,npart,iterN,seed,lambda,alpha,initial
19  format('Z,N,iterN,seed,lambda,alpha,initial ',3i4,i10,2f10.5,i3)
	read(22,*) lambdaN,lambdaP
	read(22,*) lamN_min,lamN_max
	read(22,*) lamP_min,lamP_max
	read(22,*) block_nodd, block_podd
	print*, "chemical potentials",lambdaN,lambdaP
        read(22,*) orbit_fname
        read(22,*) spshell_fname
        read(22,*) int_fname
        print*, "orbital and interaction data ",orbit_fname,spshell_fname,int_fname

	nporb = 0
        nnorb = 0
        call read_orbitals
        call read_wilden
        ! number of proton/ neutron orbits
        nnorb = norb  - nporb
        if(ppart > nporb .or. (npart ) > nnorb)then
            stop "Error: number of proton/neutron larger than number of orbits(see ufg_hw3_orbitaks.dat for number of orbits)!"
        endif
        ! check the input blocked block index is in the proper range
	if((block_nodd < 1) .or. (block_nodd > 3))then
         stop "Error: the index of the blocked neutron block is out of range!"
        endif
        if((block_podd < 4) .or. (block_podd > 6))then
         stop "Error: the index of the blocked proton block is out of range!"
        endif 
        print*, "Z,N;  ",ppart,npart
end subroutine input_data

subroutine read_orbitals
 include 'variables.dat'
  	 integer i,j,k,l
!  	 open(15,file='ufg_datas/ufg_hw3_orbitals.dat')
	 open(15,file=orbit_fname)
	 read(15,*) norb
	 print*, 'norb ',norb
  	 do i = 1,norb
     	  read(15,*) tz(i),jj2(i),jz2(i),aqn1(i)
	  if(tz(i) .eq.1)then
	   nporb = nporb + 1
	  endif		  
     	  
     	  ! block specified by projection of isospin and j 
     	  nquant(i) = (tz(i) + t_offset) * tz_offset + (jz2(i)  + m_offset)
     	  
  	 enddo
         close(15)  	 
	
end subroutine read_orbitals

subroutine block_structure
 include 'variables.dat'
	integer :: i,j,k, NB = 0, newValue, Ovalue, numE
	
	do i = 1, norb
		! select the desired orbits
		if(orbit_selector(nquant(i)) /= 1) then 
		 continue
		endif
		
		newValue = 1
		Ovalue = nquant(i)
		do j = 1, NB
			! when the orbit is in this block
			if(idx(j) == Ovalue) then
				numE = ndb(j)
				ib(numE + 1,j) = i
				ndb(j) = numE + 1
				newValue = 0
			end if
		end do
			
		! if not, construct a new block
		if(newValue == 1) then
			NB = NB + 1
			numE = 1
			ndb(NB) = numE
			ib(numE,NB) = i
			idx(NB) = Ovalue
		end if
	end do
	nblock = NB
	
 contains
  integer function orbit_selector(value)
  	integer, intent(in):: value
  	integer v1 
  	
  	
  	v1 = value / tz_offset
  	if(v1 == 0) then
  	 orbit_selector = 1
  	else
  	 orbit_selector = 0
  	endif
  end function orbit_selector 
     
end subroutine block_structure
		
!form matrix of specific interaction (used for testing)
subroutine form_q0
 include 'variables.dat'
	! mu = 0.22
	integer :: i,j,k,l, dblock, idxQ1, idxQ2
	integer :: l1,l2
	real :: Q1(4,4),sumq0
	real (8) :: clebsch
	!initialize q0
	q0 = 0.0
	
        !create <i|Q|j> matrix with i,j be p,f without spins
        Q1 = 0.0
	Q1(1,2) = -2.0 * (2**0.5)
	Q1(2,1) = Q1(1,2)
	Q1(2,2) = 2.0
	Q1(3,3) = 1.0
	Q1(4,4) = -2.0
	! start calculating q0
	call factor
	do i = 1, nblock
	 dblock = ndb(i)
	  do j = 1, dblock
	   do k = 1, dblock
	    l1 = -1
	    l2 = -1
	    ! get the l value from spherical shell
	     
	    do l=1,n_sh_max
             if((jj2(ib(j,i)) == j2sh(l)).and.(aqn1(ib(j,i)) == aqn1sh(l)))then
              l1 = lsh(l)
             endif
	     if((jj2(ib(k,i)) == j2sh(l)).and.(aqn1(ib(k,i)) == aqn1sh(l)))then
              l2 = lsh(l)
             endif
	    enddo  
	    ! transforming jm to lm+-1/2 to make <jm|Q|j'm'>
            ! first part: for spin up
	    sumq0 = 1.0
	    idxQ1 = 1 + l1/2 + abs((jz2(ib(j,i)) -1)/2)
	    idxQ2 = 1 + l2/2 + abs((jz2(ib(k,i)) -1)/2)
	    ! check if idxQ1, idxQ2 is qithin the Q1 matrix range
	    if( (idxQ1 .le. 4) .and. (idxQ2 .le. 4))then
	     sumq0 = sumq0 * Q1(idxQ1, idxQ2)
	     sumq0 = sumq0 * clebsch(2*l1,1,jj2(ib(j,i)),jz2(ib(j,i)) -1,1,jz2(ib(j,i)))
             sumq0 = sumq0 * clebsch(2*l2,1,jj2(ib(k,i)),jz2(ib(k,i)) -1,1,jz2(ib(k,i)))
	    else
             sumq0 = 0.0
	    endif
	    q0(j,k,i) = q0(j,k,i) + sumq0
	    
	    ! second part: for spin down
	    sumq0 = 1.0
	    idxQ1 = 1 + l1/2 + abs((jz2(ib(j,i)) + 1)/2)
            idxQ2 = 1 + l2/2 + abs((jz2(ib(k,i)) + 1)/2)
	    if( (idxQ1 .le. 4) .and. (idxQ2 .le. 4))then
             sumq0 = sumq0 * Q1(idxQ1, idxQ2)
             sumq0 = sumq0 * clebsch(2*l1,1,jj2(ib(j,i)),jz2(ib(j,i)) +1,-1,jz2(ib(j,i)))
             sumq0 = sumq0 * clebsch(2*l2,1,jj2(ib(k,i)),jz2(ib(k,i)) +1,-1,jz2(ib(k,i)))
            else
             sumq0 = 0.0
            endif
	    q0(j,k,i) = q0(j,k,i) + sumq0
	    
	   end do
	  end do
        end do
	! q0 times 1 over mu 
	q0 = q0 * 1.0 / mu 
end subroutine form_q0

! Create the eigenvectors of the blocks. The actual orbits will be derived from phi
! Notice that the "whole" eigenvectors are not normalized, so when you construct
! the whole eigenvector one you have to divide by the number of blocks, nblock 

! create the eigenvector for density matrix
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine form_eigenvector
 include 'variables.dat'
 	integer i,j,k,l,m
 	real (8):: fock2(ndb_MAX,ndb_MAX),z(ndb_MAX,ndb_MAX)
	real (8):: fv1(ndb_MAX),fv2(ndb_MAX),w(ndb_MAX)
	integer :: iter1,ierr
	real ::ckValue
		 
 	do l = 1,nblock
 	 ! assign fock2
 	 fock2(1:ndb(l),1:ndb(l)) = fock(1:ndb(l),1:ndb(l),l) 

	  
	 ! the diagonalizer
	 call rs(ndb_MAX,ndb(l),fock2,w,1,z,fv1,fv2,ierr)
	  
	 ! assign eigenvalues to eigenV
       	 ! Notice: eigenvalues for the same block are stored in rows
       	 eigenv(1:ndb(l),l) = w(1:ndb(l)) 
	 !print*, 'eigenvalues'
       	 !print*, (eigenv(i,l),i=1,ndb(l))
       	 !write(*,*)
       	 ! assign vectors to phi
       	 phi(1:ndb(l),1:ndb(l),l) = z 
       	 ! notice!! That the eigenvectors for the same block
       	 ! are stored in columnes!!
       	 !print*, 'vectors'
       	 !write(*,*)
       	 !do i=1,ndb(l)
         ! print*, (phi(i,k,l),k=1,ndb(l))
       	 !enddo
       	 !write(*,*)
       	 !print*, 'sanity check: should be small'
       	 call diag_check(fock2,z,w,ndb(l),ckValue)       
       	 !print*, ckValue
       	 !write(*,*) 		 
       	enddo
end subroutine form_eigenvector 	 

subroutine diag_check(fock2,z,w,ndim, ckValue) 
 include 'variables.dat'
 	integer i,j,k,l,m
 	real (8):: fock2(ndb_MAX,ndb_MAX),z(ndb_MAX,ndb_MAX)
	real (8):: fv1(ndb_MAX),fv2(ndb_MAX),w(ndb_MAX)
	real (8):: tmpVect(ndb_MAX,ndb_MAX), tmpVect1(ndb_MAX,ndb_MAX)	
	real ckValue
	integer ::ndim 
	real sumV
	
	! get the matrix
	do k=1,ndim
	 do i=1, ndim
	  sumV = 0.0
	  do j=1,ndim
	   sumV = sumV + fock2(i,j) * z(j,k)
	  enddo
	  tmpVect(i,k) = sumV
	 enddo
	enddo
	 
	do i=1,ndim
	 do j=1,ndim
	 tmpVect1(j,i) = z(j,i) * w(i)
	 enddo
	enddo
	sumV = 0.0
	do i=1,ndim
	 do j =1,ndim
	 sumV = sumV + (tmpVect(i,j) - tmpVect1(i,j))*(tmpVect(i,j) - tmpVect1(i,j))
	 enddo
	enddo
	ckValue = sumV
end subroutine diag_check

! create the density matrix
subroutine form_rho
 include 'variables.dat'
 	integer ip,in,i,j,k,l
	integer index_evn(norb),index_evp(norb) 
	integer index_ev(norb), b_ev(npart),num_bev
	real evsn(norb),evsp(norb),evs(norb),vector(norb,npart)
	!real evs(norb),vector(norb,npart)
	real tmp1
	integer tmp2,tmpo,tmpb, boffset
	
	
	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!1
	!sort the eigenvalue to find the eigenvectors 
	!corresponding to the smallest energy
	
	if(offset_ev < ndb_MAX)then
	 stop "parameters offset_ev should be bigger than ndb_MAX and be power of 10!!"
	endif
	 
	!codify the eigenvalues
	ip = 0
	in = 0

	do j=1, nblock
	 ! case: 0 is neutron; 2 is proton
	 if(idx(j) / tz_offset == 2)then
	  ! codify protons
	  do k=1,ndb(j)
	   ip = ip+1
	   index_evp(ip) = j*offset_ev + k
	   evsp(ip) = eigenv(k,j)
	   
	  enddo
	 else
	 ! otherwise codify neutrons
	  do k=1,ndb(j)
	   in = in+1
	   index_evn(in) = j*offset_ev + k
	   evsn(in) = eigenv(k,j)	   
	  enddo
	 endif
	enddo
	
	!sort indexn_ev and indexp_ev 
	!in terms of eigenvalues. 
	
	do ip=1,nporb
	 do j=ip,nporb
	  if(evsp(ip) >= evsp(j))then
	   tmp1 = evsp(ip)
	   evsp(ip) = evsp(j)
	   evsp(j) = tmp1
	   tmp2 = index_evp(ip)
	   index_evp(ip) = index_evp(j)
	   index_evp(j) = tmp2 
	  endif
	 enddo
	enddo
	
	do in=1,nnorb
	 do j=in,nnorb
	  if(evsn(in) >= evsn(j))then
	   tmp1 = evsn(in)
	   evsn(in) = evsn(j)
	   evsn(j) = tmp1
	   tmp2 = index_evn(in)
	   index_evn(in) = index_evn(j)
	   index_evn(j) = tmp2 
	  endif
	 enddo
	enddo
	
	! print out the eigenvector
	do ip=1, ppart
	 tmpb = index_evp(ip) / offset_ev
	 tmpo = index_evp(ip) - tmpb * offset_ev
	 !print*, evsp(ip)
	 !print *, (phi(j,tmpo,tmpb),j=1,ndb(tmpb))
	enddo	
	do in=1, (npart - ppart)
	 tmpb = index_evn(in) / offset_ev
	 tmpo = index_evn(in) - tmpb * offset_ev
	 !print*, evsn(in)
	 !print *, (phi(j,tmpo,tmpb),j=1,ndb(tmpb))
	enddo
	
	! finally, combine both the proton and neutron together
	do i=1,ppart
	 index_ev(i) = index_evp(i)
	 evs(i) = evsp(i)
	enddo
	do i=1, (npart - ppart)
	 index_ev(i + ppart) = index_evn(i)
	 evs(i + ppart) = evsn(i)
	enddo
	
	
	! construct the density matrix
	rho = 0.0
	do i=1,nblock
	 ! identify vectors contributing such block
	 num_bev = 0
	 do l=1,npart
	  tmp2 = index_ev(l) / offset_ev
	  if(tmp2 == i)then
	   num_bev = num_bev + 1
	   b_ev(num_bev) = l
	  endif
	 enddo
	 
	 ! calculate rho
	 do l=1,num_bev
	  tmpb = index_ev(b_ev(l)) / offset_ev
	  tmpo = index_ev(b_ev(l)) - tmpb * offset_ev  
	  do j=1,ndb(i)
	   do k=1,ndb(i)
	    rho(j,k,i) = rho(j,k,i) + phi(k,tmpo,tmpb) * &
	      phi(j,tmpo,tmpb)
	   enddo
	  enddo 
	 enddo
	enddo 	 
	
end subroutine form_rho 
	      
! create the fock matrix
subroutine form_fock
 include 'variables.dat'
	
	select case(caseidx)
	case(1)
	 call fock_case1	 
 	case(2)
 	 call fock_case2	  
 	case default
 	 call fock_general 	
 	end select	 	 
	!write(*,*)
 	!print*, "The new fock matrix"
 	!do i=1,nblock
 	! write(*,*)
 	! do j=1, ndb(i)
 	!  print*, (fock(j,k,i),k=1,ndb(i))
 	! enddo
 	!enddo	
 	
end subroutine form_fock

subroutine fock_general
 include 'variables.dat'
 integer i,j,k,l,m,n
 integer ord1, ord2
 
 real tenV
 real newfock(ndb_MAX,ndb_MAX,nblock_MAX)
 
 do i=1,nblock
  do j=1,ndb(i)
   do k=1,ndb(i)
    tenV = 0.0
    ! get the ordinary orbit indices
    ord1 = ib(j,i)
    ord2 = ib(k,i)
    ! the tensor part
    do l=1,nblock
     do m=1,ndb(l)
      do n=1,ndb(l)
       tenV = tenV + rho(m,n,l)*vv(ord1,ib(m,l),ord2,ib(n,l)) 
      enddo
     enddo
    enddo
    newfock(j,k,i) = h1(j,k,i) + lambda * q0(j,k,i)   + tenV
   
   enddo
  enddo
 enddo
 
 ! combine with the old fock to get the new fockmatrix
 fock = newfock*alpha + (1 - alpha)*fock
end subroutine fock_general
   
subroutine fock_case1
 include 'variables.dat'
 integer i,j,k,l,m
 real ten_DPart, ten_EPart,ten_sumPart
 real newfock(ndb_MAX,ndb_MAX,nblock_MAX)
 
 !for no one-body part case 
 h1 = 0.0
 
 ! case1: 
 ! v2[i][j][k][l] = -q0[i][k]*q0[j][l]
 ten_sumPart = 0.0
  do i=1,nblock
   do j=1, ndb(i)
    do k=1,ndb(i)
     ten_sumPart = ten_sumPart + q0(j,k,i)*rho(j,k,i)
    enddo
   enddo
  enddo
  ten_sumPart = -1* ten_sumPart
  ! Then get the fock matrix
 	  
  do i=1,nblock
   do j=1, ndb(i)
    do k=1,ndb(i)
 	     
     newfock(j,k,i) = h1(j,k,i) + q0(j,k,i)*ten_sumPart
    enddo
   enddo
  enddo
  
  ! combine with the old fock to get the new fockmatrix
 fock = newfock*alpha + (1 - alpha)*fock
 	
end subroutine fock_case1

subroutine fock_case2
 include 'variables.dat'
 integer i,j,k,l,m
 real ten_DPart, ten_EPart,ten_sumPart
 real newfock(ndb_MAX,ndb_MAX,nblock_MAX)
 ! case2: 
 ! v2[i][j][k][l] = q0[i][k]*q0[j][l] - q0[i][l]*q0[j][k]
 ! Note: as long as rho initially be as block-structure
 ! matrix the fock matrix constructed will remained in
 ! block-structure
 !
 ! var: ten_DPart, ten_EPart
 !
 ! the fock matrix-more general formation
 
 ! for no one-body case
 h1 = 0.0
 
 do i=1,nblock
  do j=1, ndb(i)
   do k=1, ndb(i)
    ! calculate the direct tensor part
    ten_DPart = 0.0
    do l=1, ndb(i)
      do m=1,ndb(i)
       ten_DPart = ten_DPart + q0(l,m,i)*rho(l,m,i)
      enddo
     enddo
     ! for direct part we can take this out of loop
     ten_DPart = ten_DPart * q0(j,k,i) 
   
   ! calculate the exchange tensor part
     ten_EPart = 0.0
    do l=1, ndb(i)
      do m=1,ndb(i)
       ten_EPart = ten_EPart + q0(j,m,i)*q0(l,k,i)*rho(l,m,i)
      enddo
     enddo
   
     ! then get the new fock matrix...
     newfock(j,k,i) = h1(j,k,i) + ten_DPart - ten_EPart
    enddo
   enddo
  enddo
  
  ! combine with the old fock to get the new fockmatrix
 fock = newfock*alpha + (1 - alpha)*fock
 	
end subroutine fock_case2	

! calculate the energy
subroutine get_energy
 include 'variables.dat'
 	integer i,j,k,l,m
 	
	one_body_energy = 0.0
	vv_energy = 0.0
	total_energy = 0.0
	expectQ0 = 0.0
	
	call energy_general
	call find_exp_q0 
 	
 	write(6,19) one_body_energy,vv_energy,total_energy,expectQ0
19 format("Energies:1b,2b,total,Q0 ",4f10.3)
end subroutine get_energy
	

subroutine energy_1(energy)
 include 'variables.dat'
 	integer i,j,k,l,m
 	real energy, ten_sumPart
	integer b1, b2
	energy = 0.0
	
	! case1:
 	 !  
	 ! v2[i][j][k][l] = -q0[i][k]*q0[j][l]
 	 ten_sumPart = 0.0
 	 do i=1,nblock
 	  do j=1, ndb(i)
 	   do k=1,ndb(i)
 	    ten_sumPart = ten_sumPart + q0(j,k,i)*rho(j,k,i)
 	   enddo 	  
 	  enddo
 	 enddo
 	 energy = -0.5 * ten_sumPart * ten_sumPart
 	
 	 ! plus the diagonal h1 matrix, if there is one
 	 do i=1,nblock
 	  do j=1, ndb(i)
 	   energy = energy + h1(j,j,i)*rho(j,j,i)
 	  enddo
 	 enddo
end subroutine energy_1

subroutine energy_2(energy)
 include 'variables.dat'
	integer i,j,k,l,m
 	real energy, ten_sumPart
	integer b1, b2
	energy = 0.0
	
	! case2:
  	! v2[i][j][k][l] = q0[i][k]*q0[j][l] - q0[i][l]*q0[j][k]
 	! first the direct part
 	
 	 ten_sumPart = 0.0 
 	 do i=1,nblock
 	  do j=1, ndb(i)
 	   do k=1,ndb(i)
 	    ten_sumPart = ten_sumPart + q0(j,k,i)*rho(j,k,i)
 	   enddo
 	  enddo
	 enddo
 	 energy = energy + ten_sumPart * ten_sumPart
 	  
 	 ! then the exchange part
 	
 	 ten_sumPart = 0.0
 	 do b1=1,nblock
 	  do i=1,ndb(b1)
 	   do j=1,ndb(b1)
 	    do k=1,ndb(b1)
 	     do l=1,ndb(b1)
 	      ten_sumPart = ten_sumPart + q0(i,l,b1)*q0(j,k,b1) * rho(i,k,b1) * &
 	         rho(j,l,b1)
 	     enddo
 	    enddo
 	   enddo
 	  enddo
 	 enddo
 	 energy = energy - ten_sumPart
 	 energy = 0.5 * energy
 	 
 	  ! plus the diagonal h1 matrix, if there is one
 	 do i=1,nblock
 	  do j=1, ndb(i)
 	   energy = energy + h1(j,j,i)*rho(j,j,i)
 	  enddo
 	 enddo
end subroutine energy_2

subroutine energy_general
 include 'variables.dat'
 	integer i,j,k,l,m,n
 	integer orb1, orb2
 	real e_onebody, e_tensor, tenHP1, energy
	e_onebody = 0.0
	e_tensor = 0.0
	
	! one body part
	do i=1,nblock
	 do j=1,ndb(i)
	  do k=1,ndb(i)
	   e_onebody = e_onebody + h1(j,k,i)* rho(j,k,i)
	   orb1 = ib(j,i) ! i
	   orb2 = ib(k,i) ! k
	   tenHP1 = 0.0
	   ! tensor part
	   do l=1,nblock
	    do m=1,ndb(l) ! j
	     do n=1,ndb(l) ! l
  tenHP1 = tenHP1 + vv(orb1,ib(m,l),orb2,ib(n,l))* rho(m,n,l)
	     enddo
	    enddo
	   enddo
	   e_tensor = e_tensor + tenHP1 * rho(j,k,i)/2.0
	  enddo
	 enddo
	enddo
	   
	!add both energy part together
	energy = e_tensor + e_onebody
	total_energy = energy
	one_body_energy = e_onebody
	vv_energy = e_tensor

end subroutine energy_general 
	

subroutine read_wilden
 include 'variables.dat'
  	integer i
  	real w1, w2, w3, w4
  	character (80) :: string
  	! read the shells 
!  	open(unit=7,file='ufg_datas/ufg_hw3.sps')
	open(unit=7,file=spshell_fname)
	read(7,*)n_sh_max
!	print*, "spherical shell QNumber:"
	do i=1,n_sh_max
	  read(7,*) w1,w2,w3, w4
	  nsh(i)=w1
	  lsh(i)=w2
	  j2sh(i)=2*w3
	  aqn1sh(i) = w4	  
!	  print*, nsh(i),lsh(i),j2sh(i),aqn1sh(i) 
	enddo
	! read the energy data for constructing tensor part Hamiltonian
!	open(unit=8,file='ufg_datas/ufg_hw3_q3.int')
	open(unit=8,file=int_fname)
	read(8,*) string
	read(8,*) string
	read(8,*) n_v_max,(esp(i),i=1,n_sh_max)
	do i=1,n_v_max
	  read(8,*) statek(1,i),statek(2,i),statek(3,i),statek(4,i),Jt(i),Tt(i),vw(i)
	enddo
  end subroutine read_wilden
  
  ! construct the one bosy part hamiltonian, h1
  subroutine form_h1
 include 'variables.dat'
  	integer i,j,k
  	 
  	! initialization
  	h1 = 0.0 
  	do i=1,nblock
  	 do j=1,ndb(i)
  	  h1(j,j,i) = wilden_h1(j,i)
  	 enddo
  	enddo
  	
  	contains
  	
  	real function wilden_h1(idx_inb, idx_b)
  		integer, intent(in):: idx_inb, idx_b
  		integer l_j2,l_aqn1,i
  		real energy
  		energy = 0.0
		! get orbit, then get j times 2, aqn1, then get the shell number
  		l_j2 = jj2(ib(idx_inb,idx_b))
  		l_aqn1 = aqn1(ib(idx_inb,idx_b)) 
  		
  		do i=1, n_sh_max
  		 if((l_j2 == j2sh(i)).and.(l_aqn1 == aqn1sh(i) ))then
  		   energy = esp(i)
  		 endif
  		enddo
  		wilden_h1 = energy
  	end function wilden_h1
  end subroutine form_h1
  
! Get ks from is
subroutine sm_to_orbital(is, ks)
 include 'variables.dat'
   	integer is(lVec),ks(lVec)
   	integer i,j
   	do i=1,lVec
   	 !print*,jj2(is(i))
   	 do j=1,n_sh_max   	  
   	  
	  if((jj2(is(i)) == j2sh(j)).and.(aqn1(is(i)) == aqn1sh(j)))then
   	   ks(i) = j
   	  endif
   	 enddo
   	enddo
end subroutine sm_to_orbital
! Get the tensor coefficient
subroutine vijkl
 include 'variables.dat'
  	integer i,j,k,l,m
  	integer kcoord(lVec), icoord(lVec)
  	real (8) :: sumVV, sumHP1, swapCoef
  	real (8) :: debug,clebsch
  	! stores all the states having k1,k2,k3,k4
  	integer  numState,states(n_v_max),swap_info
  	call factor
  	debug = 0.0  	
  	do i=1, norb
  	 do j=1, norb
  	  do k=1, norb
  	   do l=1, norb
  	    
  	    numState = 0
  	    icoord(1) = i
  	    icoord(2) = j
  	    icoord(3) = k
  	    icoord(4) = l
  	    
  	    
  	    ! get k1,k2,k3,k4
  	    call sm_to_orbital(icoord,kcoord)
  	    ! find the states that matches k1,k2,k3,k4

	    do m=1,n_v_max
  	     if(ck_state(kcoord,swap_info,m) == 1)then
  	      numState = numState + 1
  	      states(numState) = m
  	     endif
  	    enddo
  	    
  	    sumVV = 0.0
  	    if(numState == 0)then
  	     vv(i,j,k,l) = 0.0
  	    elseif(jz2(i) + jz2(j) /= jz2(k) + jz2(l))then
  	     vv(i,j,k,l) = 0.0
  	    elseif(tz(i) + tz(j) /= tz(k) + tz(l))then
  	     vv(i,j,k,l) = 0.0
  	    else
  	     !print*,"number of states", numState
  	     do m=1, numState
  	      ! get swapping coeficient
  	      select case(swap_info)
  	       case(0)
  	        swapCoef = 1.0
  	       case(1)
  	swapCoef = ((-1)**((jj2(k)+jj2(l))/2-Jt(states(m))+Tt(states(m))))* 1.0
  	       case(2)
  	swapCoef = ((-1)**((jj2(i)+jj2(j))/2-Jt(states(m))+Tt(states(m))))* 1.0
  	       case(3)
  	       swapCoef = ((-1)**((jj2(i)+jj2(j)+jj2(k)+jj2(l))/2)) * 1.0
  	       case default
  	       swapCoef = 0.0
  	      end select
  	      sumHP1 = vw(states(m))
  sumHP1=sumHP1*clebsch(jj2(i),jj2(j),2*Jt(states(m)),jz2(i),jz2(j),jz2(i)+jz2(j))
  sumHP1=sumHP1*clebsch(jj2(k),jj2(l),2*Jt(states(m)),jz2(k),jz2(l),jz2(k)+jz2(l))
  sumHP1=sumHP1*clebsch(1,1,2*Tt(states(m)),tz(i),tz(j),tz(i)+tz(j))
  sumHP1=sumHP1*clebsch(1,1,2*Tt(states(m)),tz(k),tz(l),tz(k)+tz(l))
  sumHP1=sumHP1*swapCoef
	
  	      
  	      sumVV = sumVV + sumHP1
  	      
  	     enddo
             if(kcoord(1).eq.kcoord(2)) then
  	       sumVV = sqrt(2.0)* sumVV 
  	     endif
             if(kcoord(3).eq.kcoord(4)) then
               sumVV = sqrt(2.0)* sumVV
             endif
	     vv(i,j,k,l) = sumVV
  	    endif
           enddo 
          enddo 
         enddo  	
  	enddo
  contains
  ! figure out what the type of k1,k2,k3,k4 is
   integer function ck_state(ks,swap_info,idx)
   	integer, intent(in):: idx
   	integer, intent(out):: swap_info
   	integer ::ks(lVec)
   	integer i,j
   	integer isState,id, base
   	base = 1
   	! construct id
   	id = id_construct(ks(1),ks(2),ks(3),ks(4)) 
if(id ==id_construct(statek(1,idx),statek(2,idx),statek(3,idx),statek(4,idx)))then
   	  ck_state = 1
   	  swap_info = 0
 elseif(id==id_construct(statek(1,idx),statek(2,idx),statek(4,idx),statek(3,idx)))then
   	  ck_state = 1
   	  swap_info = 1
 elseif(id==id_construct(statek(2,idx),statek(1,idx),statek(3,idx),statek(4,idx)))then
   	  ck_state = 1
   	  swap_info = 2
 elseif(id==id_construct(statek(2,idx),statek(1,idx),statek(4,idx),statek(3,idx)))then
   	  ck_state = 1
   	  swap_info = 3
 elseif(id==id_construct(statek(3,idx),statek(4,idx),statek(1,idx),statek(2,idx)))then
   	  ck_state = 1
   	  swap_info = 0
 elseif(id==id_construct(statek(3,idx),statek(4,idx),statek(2,idx),statek(1,idx)))then
   	  ck_state = 1
   	  swap_info = 1
 elseif(id==id_construct(statek(4,idx),statek(3,idx),statek(1,idx),statek(2,idx)))then
   	  ck_state = 1
   	  swap_info = 2
 elseif(id==id_construct(statek(4,idx),statek(3,idx),statek(2,idx),statek(1,idx)))then
   	  ck_state = 1
   	  swap_info = 3
 else
   	  ck_state = 0   	  
 endif
   end function ck_state
   
   	 
   ! construct id for determine state
   integer function id_construct(num1,num2,num3,num4)
   	integer, intent(in):: num1, num2, num3, num4
   	id_construct = 1000 * num1 + 100 * num2 + 10 * num3 + num4
   end function id_construct
end subroutine vijkl

! gives Clebsch Gorden coef. The argument is ALL TWO times the acyual Q number!! 
function CLEBSCH(J1,J2,J3,M1,M2,M3)
  integer, intent(in) :: J1, J2, J3, M1, M2, M3
  real (8) :: IA,IB,IC,ID,IE,IF,IG, IH,II, IJ
  real (8) :: NUMIN1, NUMAX1,KB, KC, NUM1, NU         
  real (8) :: faclog(130),first,part1,second,con,ff,clebsch
  
  COMMON/BL1/FACLOG
  IF(M1+M2.NE.M3)GO TO 130
  if(abs(M3) > J3) go to 130 
      IF(J3.LE.J1+J2.AND.J3.GE.ABS (J1-J2))GO TO 110
      GO TO 130
  110 IA=(J1+J2-J3+2)/2.0
      IB=(J3+J1-J2+2)/2.0
      IC=(J3+J2-J1+2)/2.0
      ID=(J1+J2+J3+4)/2.0
      IE=(J1+M1+2)/2.0
      IF=(J1-M1+2)/2.0
      IG=(J2+M2+2)/2.0
      IH=(J2-M2+2)/2.0
      II=(J3+M3+2)/2.0
      IJ=(J3-M3+2)/2.0
      FIRST=0.5*(LOG(J3+1.0)+FACLOG(IA)+FACLOG(IB)+FACLOG(IC)+FACLOG(IE)+FACLOG(IF)+FACLOG(IG)+FACLOG(IH)+FACLOG(II)+FACLOG(IJ)-FACLOG(ID))
      PART1=EXP (FIRST)
      NUMIN1= ABS(AMIN1(((J3-J2+M1)/2.0),((J3-J1-M2)/2.0),0.0))+1.0
      NUMAX1=MIN1(((J1+J2-J3)/2.0),((J1-M1)/2.0),((J2+M2)/2.0))+1
      KB=(J3-J2+M1)/2.0+1.0
      KC=(J3-J1-M2)/2.0+1.0
      FF=0
      DO 120  NUM1=NUMIN1,NUMAX1
      NU=NUM1-1
      SECOND=-(FACLOG(NUM1)+FACLOG(IA-NU)+FACLOG(IF-NU)+FACLOG(IG-NU)+FACLOG(KB+NU)+FACLOG(KC+NU))
      CON=((-1)**NU)*EXP (SECOND)
      FF=FF+CON
  120 CONTINUE
      clebsch=PART1*FF
      GO TO 140
  130 clebsch=0
  140 CONTINUE
      RETURN 
END function CLEBSCH
   
! subroutine needed to be called before using clebsch function
subroutine factor
  implicit none
  integer M
  real (8) ::faclog(130),fn
  COMMON/BL1/FACLOG
  
  FACLOG(1)=0
  FACLOG(2)=0
  FN=1
  DO   M=3,130
      FN=FN+1
      FACLOG(M)=FACLOG(M-1)+LOG(FN)
  enddo
  RETURN
END subroutine factor	  		
 
 
! A very-very simple initial fock matrix generator
subroutine init_fock
 include 'variables.dat'
 	integer i,j,k,l,m
 	integer tarb, num
	tarb = 3
	
		
	fock = 0.0
	do i=1,nblock
	 do j=1,ndb(i)
	  if(i == tarb)then
	   fock(j,j,i) = -1.0
	   
	  else
	   fock(j,j,i) = 1.0
	  endif
	 enddo
	enddo
end subroutine init_fock
	  
! one using random number generator
subroutine init_fock_r
 include 'variables.dat'
    	integer i,j,k,l,m
    	real tmpv1
  	
  	fock = 0.0
  	
  	do i=1,nblock
	 do j=1,ndb(i)
	  do k=1,j
	   tmpv1 = (0.5 - ran(seed)) * 2.0
	   fock(j,k,i) = tmpv1
	   fock(k,j,i) = fock(j,k,i)
	   
	  enddo
	 enddo
	enddo
end subroutine init_fock_r
  	
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!11
!!!!The followings are subroutines specifically for hfb functions!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

! initialize fock and deltam matrixes
subroutine hfb_init_fock_deltam
 include 'variables.dat'
    	integer i,j,k,l,m
    	real tmpv1
  	
  	fock = 0.0
	deltam = 0.0
	
  	
!  	do i=1,nblock
!	 do j=1,ndb(i)
!	  do k=1,j
!	   tmpv1 = (0.5 - ran(seed)) * 2.0
!	   fock(j,k,i) = tmpv1
!	   fock(k,j,i) = fock(j,k,i)
!	   
!	  enddo
!	 enddo
!	enddo
        do i=1,hfb_nb
         do j=1,ndb(hfb_block(i,1))
          do k=1,j
           tmpv1 = (0.5 - ran(seed)) * 2.0
           fock(j,k,hfb_block(i,1)) = tmpv1
           fock(j,k,hfb_block(i,2)) = tmpv1
           fock(k,j,hfb_block(i,1)) = fock(j,k,hfb_block(i,1))
           fock(k,j,hfb_block(i,2)) = fock(j,k,hfb_block(i,2))

          enddo
         enddo
        enddo




	
	do i=1,hfb_nb
	 do j=1,ndb(hfb_block(i,1))
	  do k=1,ndb(hfb_block(i,1))
	   tmpv1 = (0.5 - ran(seed)) * 2.0
	   deltam(j,k,i) = tmpv1
	   deltam(k,j,i) = deltam(j,k,i)
	   
	  enddo
	 enddo
	enddo	
end subroutine hfb_init_fock_deltam

subroutine hfb_block_structure
! requires block_structure initialization
 include 'variables.dat'
 	integer i,j, used(nblock),it,t_upper,t_lower
	integer jz_upper, jz_lower, iter1
	used = 0
	iter1 = 0
!  second block is chosen by index2 = -index1.  Set index to 0 to duplicate.
	do i=1,nblock	 
	 if(used(i) .eq. 0)then
! extract jz from idx
	  jz_upper = mod(idx(i),tz_offset) - m_offset
          t_upper = idx(i)/tz_offset 
          if (jz_upper.ge.0) then
!          print*, i, jz_upper
	  do j=1,nblock
	   jz_lower = mod(idx(j),tz_offset) - m_offset
           t_lower = idx(j)/tz_offset 
           if(jz_lower.eq.-jz_upper.and.t_upper.eq.t_lower) then
	    iter1 = iter1 + 1
	    ! assign index for hfb block
	    hfb_block_quant(iter1) = jz_upper
	    hfb_block(iter1,1) = i
	    hfb_block(iter1,2) = j
	    used(i) = j
	    used(j) = i
	   endif
	  enddo
          endif
	 endif
	enddo
	hfb_nb = iter1   ! number of hfb block
        print*, 'number of hfb blocks ',hfb_nb
end subroutine hfb_block_structure


subroutine hfb_construct
 include 'variables.dat'
 integer i,j,k,b1,b2, nd
 real lambdaNP
 do i=1,hfb_nb
  b1 = hfb_block(i,1)
  b2 = hfb_block(i,2)
  nd = ndb(b1)
  ! assign hfb_matrix --- fock part
  if( (idx(b1) / tz_offset) .gt. 1)then
   lambdaNP = lambdaP
  else
   lambdaNP = lambdaN
  endif
  
  do k=1,nd
    hfb_matrix(1:nd,k,i) = fock(1:nd,k,b1)
    hfb_matrix(k,k,i) = hfb_matrix(k,k,i) - lambdaNP   
    hfb_matrix(nd+1:2*nd,k+nd,i) = -1.0*fock(1:nd,k,b2)
    hfb_matrix(k + nd,k + nd,i) = hfb_matrix(k + nd,k + nd,i) + lambdaNP 
    hfb_matrix(k+nd,1:nd,i) = deltam(1:nd,k,i)
    hfb_matrix(1:nd,k+nd,i) = deltam(1:nd,k,i)
  enddo
  do k=1,2*nd
!    write(6,12) (hfb_matrix(k,j,i),j=1,2*nd)
12  format(8f9.4)
  enddo
 enddo
end subroutine hfb_construct



subroutine hfb_diag1
! diagonalize, extract positive eigensolutions
 include 'variables.dat'
 integer i,ndim,ierr
 real (8):: matrix(hfb_ndb_MAX,hfb_ndb_MAX),z(hfb_ndb_MAX,hfb_ndb_MAX)
 real (8):: fv1(hfb_ndb_MAX),fv2(hfb_ndb_MAX),w(hfb_ndb_MAX)
 do i=1,hfb_nb
  ndim = ndb(hfb_block(i,1)) * 2 
  matrix(1:ndim,1:ndim) = hfb_matrix(1:ndim,1:ndim,i)
  call rs(hfb_ndb_MAX,ndim,matrix,w,1,z,fv1,fv2,ierr)
  ! assign eigenvalues
  hfb_eigenv(1:ndim,i) = w(1:ndim)
  ! assign uvs
  hfb_uvs(1:ndim,1:ndim,i) = z
 enddo
end subroutine hfb_diag1
   
subroutine new_rho_kappa_e
 include 'variables.dat'
 integer :: i,k1,k2,b1,b2,m,ndim
 real :: rho1_sum, kappa1_sum, rho2_sum, kappa2_sum
 ! loop through the HFB blocks
 converge_rho = 0.0
 converge_kappa = 0.0
 do i=1, hfb_nb
   ! index of the hf block
   b1 = hfb_block(i,1)
   b2 = hfb_block(i,2)
   ndim=ndb(b1)
   do k1=1,ndim
    do k2=1,ndim
     rho1_sum = 0.0
     kappa1_sum = 0.0
     rho2_sum = 0.0
     kappa2_sum = 0.0
     do m=1, 2*ndim
      if(hfb_eigenv(m,i) >= 0.0)then
       rho1_sum = rho1_sum + hfb_uvs(ndim+  k1,m,i) * &
              hfb_uvs(k2+ndim ,m,i)
       kappa1_sum = kappa1_sum + hfb_uvs(k1,m,i) * hfb_uvs(k2+ndim,m,i)
      else
       rho2_sum = rho2_sum + hfb_uvs(k1,m,i)* hfb_uvs(k2,m,i)
       kappa2_sum = kappa2_sum + hfb_uvs(k2,m,i) * hfb_uvs(k1+ndim,m,i)        
      endif       
     enddo
     rho(k1,k2,b1) = rho1_sum
     kappa(k1,k2,b1) = -kappa1_sum
     rho(k1,k2,b2) = rho2_sum
     kappa(k1,k2,b2) = -kappa2_sum
     
    enddo
   enddo 
 enddo


   
end subroutine new_rho_kappa_e

subroutine new_rho_kappa_o(blko)
 include 'variables.dat'
 integer :: i,k1,k2,b1,b2,m,ndim,blko,m0
 real :: rho1d,rho2d,kappad,eig0,eig
 ! correct blko for odd number of particles
   b1 = hfb_block(blko,1)
   b2 = hfb_block(blko,2)
   ndim=ndb(b1)
 ! find highest negative eigenvalue
   m0=0
   eig0=-100.0
!   print*, hfb_eigenv(1:2*ndim, blko)
   do m=1,2*ndim
     eig=hfb_eigenv(m,blko)
     if( eig < 0.0 ) then
       if ( eig > eig0 ) then
          m0 = m
          eig0=eig
       endif
     endif
   enddo
! found u3, v3
!   print*, m0, eig0
   do k1=1,ndim
    do k2=1,ndim
     rho1d =  hfb_uvs(ndim+  k1,m0,blko) * &
              hfb_uvs(k2+ndim ,m0,blko)
     kappad =  hfb_uvs(k1,m0,blko) * hfb_uvs(k2+ndim,m0,blko)
     rho(k1,k2,b1) = rho(k1,k2,b1) + rho1d
     kappa(k1,k2,b1) = kappa(k1,k2,b1) - kappad
     kappa(k1,k2,b2) = kappa(k1,k2,b2) + hfb_uvs(k1+ndim,m0,blko) * hfb_uvs(k2,m0,blko) 
!     kappa(k1,k2,b2) = -kappa(k1,k2,b1)
     rho2d = hfb_uvs(k1,m0,blko) * hfb_uvs(k2,m0,blko)
     rho(k1,k2,b2) = rho(k1,k2,b2) - rho2d     
    enddo
   enddo 
end subroutine new_rho_kappa_o







subroutine new_delta
 include 'variables.dat'
 integer i,i1,i2,ib1,ib2,ndim1,k,k1,k2,kb1,kb2,ndim2
 real deltam_sum
 real newdeltam(ndb_MAX,ndb_MAX,nblock_MAX)
 do i=1,hfb_nb
  ib1 = hfb_block(i,1)
  ib2 = hfb_block(i,2)
  ndim1=ndb(ib1)
  do i1=1,ndim1
   do i2=1,ndim1
       deltam_sum = 0.0
       do k=1,hfb_nb
        kb1=hfb_block(k,1)
        kb2 = hfb_block(k,2)
        ndim2= ndb(kb1)
        do k2=1,ndim2
         do k1=1,ndim2
          deltam_sum = deltam_sum + kappa(k1,k2,kb1)*vv(ib(i1,ib1),ib(i2,ib2),ib(k1,kb1),ib(k2,kb2))
         enddo
        enddo
       enddo
       newdeltam(i1,i2,i) = deltam_sum
   enddo
  enddo
 enddo
 deltam = newdeltam * alpha + (1.0 - alpha)*deltam
end subroutine new_delta

subroutine new_fock
! notice this fock includes the quadrupole field
 include 'variables.dat'
 integer i,j,k
 real :: hfb_lambda
 call fock_general
 ! adding the chemical potential lambda after 
 ! the new fock matrix is mixed with previous fock matrix 
 do i=1,nblock
  if(idx(i)/tz_offset.le.t_offset) then
    hfb_lambda=lambdaN
  else
    hfb_lambda=lambdaP
  endif
  do j=1,ndb(i)
   fock(j,j,i) = fock(j,j,i) - hfb_lambda
  enddo
 enddo
end subroutine new_fock



subroutine hfb_get_energy
 include 'variables.dat'
 integer i,j,k,l,m,n,b1,b2,i1,i2,j1, jb1,jb2,nd,j2,nd2
 integer orb1, orb2
 real e_onebody,  tenHP1,e_pairingHP1, energy,temp,t2,t3
 one_body_energy = 0.0
 vv_energy = 0.0
 total_energy = 0.0
 
 e_onebody = 0.0
 e_pairing = 0.0
 
 ! loop through hf
 do i=1,nblock
  do j=1,ndb(i)
   do k=1,ndb(i)
    e_onebody = e_onebody + h1(j,k,i)* rho(j,k,i)
    orb1 = ib(j,i) ! i
    orb2 = ib(k,i) ! k
    tenHP1 = 0.0
    t2 = 0.0
    do l=1,nblock
     do m=1,ndb(l) ! j
      do n=1,ndb(l) ! l
   tenHP1 = tenHP1 + vv(orb1,ib(m,l),orb2,ib(n,l))* rho(m,n,l) !ijkl
      enddo
     enddo
    enddo
    vv_energy = vv_energy  + tenHP1 * rho(j,k,i)/2.0
   enddo
  enddo
 enddo

 do i = 1, hfb_nb
   b1=hfb_block(i,1)
   b2 = hfb_block(i,2)
   nd = ndb(b1)
   do i1=1,nd
     do i2 = 1,nd
       t2=0.0
       t3=0.0
       do j = 1, hfb_nb
         jb1= hfb_block(j,1)
         jb2= hfb_block(j,2)
         nd2=ndb(jb1)
         do j1=1,nd2
           do j2 = 1,nd2

       t2 = t2  + vv(ib(i1,b1),ib(i2,b2),ib(j1,jb1),ib(j2,jb2) ) &
         * kappa(j1,j2,jb1)
       t3 = t3  + vv(ib(i1,b1),ib(i2,b2),ib(j1,jb1),ib(j2,jb2) ) &
         * kappa(j1,j2,jb2)
           enddo
         enddo
       enddo
       e_pairing = e_pairing + (t2*kappa(i1,i2,b1) + t3*kappa(i1,i2,b2))*0.5  
        
     enddo
   enddo
 enddo


 one_body_energy = e_onebody
 total_energy = vv_energy  + e_onebody + e_pairing
 
  write(6,19) one_body_energy,vv_energy,e_pairing, total_energy
19 format("Energies:1b,2bhf,2bpairing,total ",5f10.3)

end subroutine hfb_get_energy 

subroutine read_kappa
include 'variables.dat'
integer :: i,j,k,ndim
!  kappa must be antisymmetric.  therefore blocks  (4,5,6) = - (1,2,3) and 
!                                                  (10,11,12) = - (7,8,9)
open(14)
  do i=1,nblock
    ndim=ndb(i)
    do j=1,ndim
      read(14,*) (kappa(k,j,i),k=1,ndim)
    enddo
  enddo
close(14)
end subroutine read_kappa
  
subroutine write_kappa
include 'variables.dat'
integer :: i,j,k,ndim
  do i=1,nblock
    ndim=ndb(i)
    do j=1,ndim
      write(15,14) (kappa(k,j,i),k=1,ndim)
14 format(8f10.5)
    enddo
  enddo
end subroutine write_kappa
        
subroutine find_numbers(Ncal,Zcal)
include 'variables.dat'
  integer :: i,j
  real :: Ncal, Zcal
  Ncal=0.0
  Zcal=0.0
  do i=1,nblock
    do j=1,ndb(i)
      if(idx(i)/tz_offset.le.t_offset) then
         Ncal=Ncal+rho(j,j,i)
      else
         Zcal=Zcal+rho(j,j,i)
      endif
    enddo
   enddo
end subroutine find_numbers    

subroutine form_odd_number_rho_kappa
 include 'variables.dat'
 integer :: i,j,k,l,ndim
 integer :: minidxb, minidxev
 real :: minev 
 ! do neutron first
 if(mod(npart,2) .eq. 1) then
  print*, "Modify neutron rho for odd-number neutron"
  ! find the lowest positive eigenvalues
  ! and the corresponding eigenvectors
  minidxb = 0
  minidxev = 0 
  minev = 1.0e12
  do i=1,(hfb_nb/2)
   ndim = ndb(hfb_block(i,1))
   do j=1,2*ndim
    if( (hfb_eigenv(j,i) > 0.0) .and. (hfb_eigenv(j,i) < minev))then
     minidxev=j
     minidxb=i
    endif
   enddo
  enddo
  ! modify the corresponding rho matrix
  ndim=ndb(hfb_block(minidxb,1))
  do i=1,ndim
   do j=1,ndim
    ! 1. The rho with Jz >0
    rho(i,j,hfb_block(minidxb,1)) = rho(i,j,hfb_block(minidxb,1)) &
     + hfb_uvs(i,minidxev,minidxb) * hfb_uvs(j,minidxev,minidxb)
    ! 2. The rho with Jz <0
    rho(i,j,hfb_block(minidxb,2)) = rho(i,j,hfb_block(minidxb,2)) &
     - hfb_uvs(i,minidxev,minidxb) * hfb_uvs(j,minidxev,minidxb)
    ! 3. kappa with Jz >0
    kappa(i,j,hfb_block(minidxb,1)) = kappa(i,j,hfb_block(minidxb,1)) &
     + hfb_uvs(ndim + i,minidxev,minidxb) * hfb_uvs(j,minidxev,minidxb)
    ! 4. kappa with Jz <0
    kappa(i,j,hfb_block(minidxb,2)) = kappa(i,j,hfb_block(minidxb,2)) &
     - hfb_uvs(ndim + i,minidxev,minidxb) * hfb_uvs(j,minidxev,minidxb)    
    
   enddo
  enddo
   

  
 endif
 
 ! Then proton
 if(mod(ppart,2) .eq. 1) then
  print*, "Modify proton rho for odd-number proton"
  ! find the lowest positive eigenvalues
  ! and the corresponding eigenvectors
  minidxb = 0
  minidxev = 0
  minev = 1.0e12
  do i=(hfb_nb/2) + 1, hfb_nb
   ndim = ndb(hfb_block(i,1))
   do j=1,2*ndim
    if( (hfb_eigenv(j,i) > 0.0) .and. (hfb_eigenv(j,i) < minev))then
     minidxev=j
     minidxb=i
    endif
   enddo
  enddo
  ! modify the corresponding rho vector
  ndim=ndb(hfb_block(minidxb,1))
  do i=1,ndim
   do j=1,ndim
    ! 1. The rho with Jz >0
    rho(i,j,hfb_block(minidxb,1)) = rho(i,j,hfb_block(minidxb,1)) &
     + hfb_uvs(i,minidxev,minidxb) * hfb_uvs(j,minidxev,minidxb)
    ! 2. The rho with Jz <0
    rho(i,j,hfb_block(minidxb,2)) = rho(i,j,hfb_block(minidxb,2)) &
     - hfb_uvs(i,minidxev,minidxb) * hfb_uvs(j,minidxev,minidxb)
    ! 3. kappa with Jz >0
    kappa(i,j,hfb_block(minidxb,1)) = kappa(i,j,hfb_block(minidxb,1)) &
     + hfb_uvs(ndim + i,minidxev,minidxb) * hfb_uvs(j,minidxev,minidxb)
    ! 4. kappa with Jz <0
    kappa(i,j,hfb_block(minidxb,2)) = kappa(i,j,hfb_block(minidxb,2)) &
     - hfb_uvs(ndim + i,minidxev,minidxb) * hfb_uvs(j,minidxev,minidxb)
   enddo
  enddo

  ! modify the corresponding kappa vector
 endif
end subroutine form_odd_number_rho_kappa
  
   
     
